
#include "PicoInt.h"

int (*PicoScan)(unsigned int num,unsigned short *data)=NULL;

static unsigned short ScanData[336]; // Scanline + messy border
static int Scanline=0; // Scanline

static int CramToHigh(int cram)
{
  int high=0x18e3;
  // Convert 0000bbb0 ggg0rrr0
  // to      rrr11ggg 111bbb11
  high|=(cram&0x00e)<<12; // Red
  high|=(cram&0x0e0)<< 3; // Green
  high|=(cram&0xe00)>> 7; // Blue
  return high;
}

static void TileNorm(unsigned short *pd,int addr,unsigned short *pal)
{
  unsigned int pack=0; int t=0,n=0;

  pack=(Pico.vram[addr]<<16)|Pico.vram[addr+1]; // Get 8 pixels

  if (pack) for (n=0; n<=7; n++,pack<<=4) { t=pack>>28; if (t) pd[n]=pal[t]; }
}

static void TileFlip(unsigned short *pd,int addr,unsigned short *pal)
{
  unsigned int pack=0; int t=0,n=0;

  pack=(Pico.vram[addr]<<16)|Pico.vram[addr+1]; // Get 8 pixels

  if (pack) for (n=7; n>=0; n--,pack<<=4) { t=pack>>28; if (t) pd[n]=pal[t]; }
}

static int DrawLayer(int plane,int high)
{
  struct PicoVideo *pvid=&Pico.video;
  int htab=0;
  int vscroll=0,hscroll=0;
  int nametab=0;
  int width=0,height=0;
  int xmask=0,ymask=0;
  int line=0;
  char shift[4]={5,6,6,7};
  int tilex=0,dx=0,ty=0;

  // Get vertical scroll value:
  vscroll=Pico.vsram[plane];

  htab=pvid->reg[13]<<9; // Horizontal scroll table address
  if ( pvid->reg[11]&2)     htab+=Scanline<<1; // Offset by line
  if ((pvid->reg[11]&1)==0) htab&=~0xf; // Offset by tile
  htab+=plane; // A or B

  // Get horizontal scroll value
  hscroll=Pico.vram[htab&0x7fff];

  // Work out the name table size: 32 64 or 128 tiles (0-3)
  width=pvid->reg[16];
  height=(width>>4)&3;
  width&=3;

  xmask=(1<<shift[width ])-1; // X Mask in tiles
  ymask=(8<<shift[height])-1; // Y Mask in pixels

  // Find name table:
  if (plane==0) nametab=(pvid->reg[2]&0x38)<< 9; // A
  else          nametab=(pvid->reg[4]&0x07)<<12; // B

  // Find the line in the name table
  line=(vscroll+Scanline)&ymask;
  nametab+=(line>>3)<<shift[width];

  // Draw tiles across screen:
  tilex=(-hscroll)>>3;
  ty=(line&7)<<1; // Y-Offset into tile
  for (dx=((hscroll-1)&7)+1; dx<328; dx+=8,tilex++)
  {
    int code=0,addr=0;
    unsigned short *pal=NULL;

    code=Pico.vram[nametab+(tilex&xmask)];
    if ((code>>15)!=high) continue;

    // Get tile address/2:
    addr=(code&0x7ff)<<4;
    if (code&0x1000) addr+=14-ty; else addr+=ty; // Y-flip

    pal=Pico.highpal+((code>>9)&0x30); // Get palette pointer

    if (code&0x0800) TileFlip(ScanData+dx,addr,pal);
    else             TileNorm(ScanData+dx,addr,pal);
  }

  return 0;
}

static int DrawSprite(int sy,unsigned short *sprite,int high)
{
  int sx=0,width=0,height=0;
  int row=0,code=0;
  unsigned short *pal=NULL;
  int tile=0,delta=0;
  int i=0;

  code=sprite[2];
  if ((code>>15)!=high) return 0; // Wrong priority

  height=sprite[1]>>8;
  width=(height>>2)&3; height&=3;
  width++; height++; // Width and height in tiles
  if (Scanline>=sy+(height<<3)) return 0; // Not on this line after all

  row=Scanline-sy; // Row of the sprite we are on
  pal=Pico.highpal+((code>>9)&0x30); // Get palette pointer
  if (code&0x1000) row=(height<<3)-1-row; // Flip Y

  tile=code&0x7ff; // Tile number
  tile+=row>>3; // Tile number increases going down
  delta=height; // Delta to increase tile by going right
  if (code&0x0800) { tile+=delta*(width-1); delta=-delta; } // Flip X

  tile<<=4; tile+=(row&7)<<1; // Tile address
  delta<<=4; // Delta of address

  sx=(sprite[3]&0x3ff)-0x78; // Get X coordinate + 8

  for (i=0; i<width; i++,sx+=8,tile+=delta)
  {
    if (sx<=0 || sx>=328) continue; // Offscreen

    tile&=0x7fff; // Clip tile address
    if (code&0x0800) TileFlip(ScanData+sx,tile,pal);
    else             TileNorm(ScanData+sx,tile,pal);
  }

  return 0;
}

static int DrawAllSprites(int high)
{
  struct PicoVideo *pvid=&Pico.video;
  int table=0;
  int i=0,link=0;

  table=pvid->reg[5]&0x7f;
  if (pvid->reg[12]&1) table&=0x7e; // Lowest bit 0 in 40-cell mode
  table<<=8; // Get sprite table address/2

  for (i=0;i<80;i++)
  {
    unsigned short *sprite=NULL;
    int sy=0;
    
    sprite=Pico.vram+((table+link)&0x7ffc); // Find sprite
    sy=(sprite[0]&0x3ff)-0x80; // Get Y coordinate

    if (Scanline>=sy && Scanline<sy+32) DrawSprite(sy,sprite,high); // Possibly on this line

    // Find next sprite
    link=sprite[1]&0x7f;
    if (link==0) break; // End of list
    link<<=2;
  }

  return 0;
}

int PicoDraw()
{
  int c=0,x=0;

  if (PicoScan==NULL) return 1; // Not drawing image

  // Convert palette:
  for (c=0;c<64;c++) Pico.highpal[c]=(unsigned short)CramToHigh(Pico.cram[c]);

  // Draw screen:
  for (Scanline=0; Scanline<224; Scanline++)
  {
    unsigned short *pd=NULL;
    unsigned int back=0;

    // Start with a blank scanline (background colour):
    back=Pico.video.reg[7]&0x3f;
    back=Pico.highpal[back];
    for (x=8;x<328;x++) ScanData[x]=(unsigned short)back;

    if (Pico.video.reg[1]&0x40)
    {
      // Draw display:
      DrawLayer(1,0);
      DrawLayer(0,0);
      DrawAllSprites(0);
      DrawLayer(1,1);
      DrawLayer(0,1);
      DrawAllSprites(1);
    }

    pd=ScanData+280-68;

    PicoScan(Scanline,ScanData+8);
  }

  return 0;
}
